import os
import sys
from urllib.request import urlopen

from hatchling.builders.hooks.plugin.interface import BuildHookInterface

JUPYTERLAB_APPUTILS_VERSION = "3.2.8"
JUPYTERLAB_THEME_LIGHT_VERSION = "3.2.8"

CSS_FILES = [
    (
        f"https://unpkg.com/@jupyterlab/apputils@{JUPYTERLAB_APPUTILS_VERSION}/style/materialcolors.css",
        "materialcolors.css",
    ),
    (
        f"https://unpkg.com/@jupyterlab/theme-light-extension@{JUPYTERLAB_THEME_LIGHT_VERSION}/style/variables.css",
        "labvariables.css",
    ),
]


class CustomBuildHook(BuildHookInterface):
    def initialize(self, version, build_data):
        for url, filename in CSS_FILES:
            directory = os.path.join(
                "share", "jupyter", "voila", "templates", "base", "static"
            )
            dest = os.path.join(directory, filename)
            if not os.path.exists(directory):
                os.makedirs(directory)
            if not os.path.exists(".git") and os.path.exists(dest):
                # not running from git, nothing to do
                return
            print("Downloading CSS: %s" % url)
            try:
                css = urlopen(url).read()
            except Exception as e:
                msg = f"Failed to download css from {url}: {e}"
                print(msg, file=sys.stderr)

                if os.path.exists(dest):
                    print("Already have CSS: %s, moving on." % dest)
                else:
                    raise OSError("Need CSS to proceed.")
                return

            with open(dest, "wb") as f:
                f.write(css)
            print("Downloaded Notebook CSS to %s" % dest)



import os

import pytest

NOTEBOOK_PATH = "syntax_error.ipynb"


@pytest.fixture
def voila_notebook(notebook_directory):
    return os.path.join(notebook_directory, NOTEBOOK_PATH)


async def test_syntax_error(http_server_client, base_url):
    response = await http_server_client.fetch(base_url)
    assert response.code == 200
    output = response.body.decode("utf-8")
    assert "There was an error when executing cell" in output
    assert "This should not be executed" not in output



import os

import pytest


@pytest.fixture
def voila_notebook(notebook_directory):
    return os.path.join(notebook_directory, "autokill.ipynb")


@pytest.fixture
def voila_args_extra():
    return ["--debug"]


@pytest.fixture
def preheat_mode():
    return False


async def test_kernel_death(http_server_client, base_url):
    response = await http_server_client.fetch(base_url)
    html_text = response.body.decode("utf-8")
    assert "raise DeadKernelError" in html_text



# test basics of Voilà running a notebook


async def test_hello_world(http_server_client, print_notebook_url):
    response = await http_server_client.fetch(print_notebook_url)
    assert response.code == 200
    html_text = response.body.decode("utf-8")
    assert "Hi Voilà" in html_text
    assert "print(" not in html_text, "by default the source code should be stripped"
    assert (
        "test_template.css" not in html_text
    ), "test_template should not be the default"



import os

# Add dev disclaimer.
_release = {}
exec(
    compile(open("../voila/_version.py").read(), "../voila/_version.py", "exec"),
    _release,
)
if _release["version_info"][-1] == "dev":
    rst_prolog = """
    .. note::

        This documentation is for a development version of Voilà. There may be
        significant differences from the latest stable release.

    """

on_rtd = os.environ.get("READTHEDOCS", None) == "True"

html_theme = "pydata_sphinx_theme"
html_theme_options = {"github_url": "https://github.com/voila-dashboards/voila"}


extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinxcontrib.video",
    "myst_parser",
]

myst_enable_extensions = ["colon_fence"]
source_suffix = ".md"
master_doc = "index"
project = "voila"
copyright = "2020, The Voilà Development Team"
author = "The Voilà Development Team"
version = ".".join(map(str, _release["version_info"][:2]))
release = _release["__version__"]
language = "en"

html_logo = "voila-logo.svg"

exclude_patterns = []
highlight_language = "python"
pygments_style = "sphinx"
todo_include_todos = False
htmlhelp_basename = "voiladoc"

intersphinx_mapping = {"python": ("https://docs.python.org", None)}

myst_heading_anchors = 3



# tests programmatic config of template system
import os

import pytest

BASE_DIR = os.path.dirname(__file__)


@pytest.fixture
def voila_args_extra():
    return ["--template=test_template", "--VoilaExecutor.timeout=240"]


@pytest.fixture
def voila_config():
    def config(app):
        path_test_template = os.path.abspath(
            os.path.join(
                BASE_DIR, "../test_template/share/jupyter/voila/templates/test_template"
            )
        )
        path_default = os.path.abspath(
            os.path.join(BASE_DIR, "../../share/jupyter/voila/templates/default")
        )
        app.template_paths = [path_test_template, path_default]

    return config


async def test_template(http_server_client, base_url):
    response = await http_server_client.fetch(base_url)
    assert response.code == 200
    assert "test_template.css" in response.body.decode("utf-8")
    assert "Hi Voilà" in response.body.decode("utf-8")



#############################################################################
# Copyright (c) 2018, Voilà Contributors                                    #
# Copyright (c) 2018, QuantStack                                            #
#                                                                           #
# Distributed under the terms of the BSD 3-Clause License.                  #
#                                                                           #
# The full license is in the file LICENSE, distributed with this software.  #
#############################################################################

import os
import re

import tornado.web

from .paths import collect_static_paths


class TemplateStaticFileHandler(tornado.web.StaticFileHandler):
    """Static file handler that serves the static files for the template system.

    URL paths should be of the form <`template_name>/static/<path>`

    A url such as lab/static/voila.js can be translated to a real path such
    /my/prefix/jupyter/voila/templates/base/static/voila.js
    Meaning the url portion is not part of the real (absolute path)

    For this system, we don't need to use the root, since this is handled in the
    paths module.
    """

    def initialize(self):
        super().initialize(path="/fake-root/voila-template-system/")

    def parse_url_path(self, path):
        # since we cannot determine the template from validate_absolute_path
        # we get all possible roots here:
        template, static, _ignore = path.split("/", 2)
        assert static == "static"
        self.roots = collect_static_paths(["voila", "nbconvert"], template)
        # simply forward the call
        return super().parse_url_path(path)

    def validate_absolute_path(self, root: str, absolute_path: str):
        # Instead of comparing to 1 root (self.root), which we don't use
        # we compare it to all the roots, and only 1 combinations has to be valid
        last_exception = None
        for root in self.roots:
            try:
                return super().validate_absolute_path(root, absolute_path)
            except tornado.web.HTTPError as e:
                last_exception = e
        assert last_exception
        raise last_exception

    @classmethod
    def get_absolute_path(cls, root, path):
        template, static, relpath = os.path.normpath(path).split(os.path.sep, 2)
        assert static == "static"
        roots = collect_static_paths(["voila", "nbconvert"], template)
        for root in roots:
            abspath = os.path.abspath(os.path.join(root, relpath))
            if os.path.exists(abspath):
                return abspath
                break
        # if we didn't find it, we will simply return an invalid path
        # which will lead to a 404
        return abspath


class MultiStaticFileHandler(tornado.web.StaticFileHandler):
    """A static file handler that 'merges' a list of directories

    If initialized like this::

        application = web.Application([
            (r"/content/(.*)", web.MultiStaticFileHandler, {"paths": ["/var/1", "/var/2"]}),
        ])

    A file will be looked up in /var/1 first, then in /var/2.

    """

    def initialize(self, paths, default_filename=None):
        self.roots = paths
        super().initialize(path=paths[0], default_filename=default_filename)

    def get_absolute_path(self, root, path):
        # find the first absolute path that exists
        self.root = self.roots[0]
        for root in self.roots:
            abspath = os.path.abspath(os.path.join(root, path))
            if os.path.exists(abspath):
                self.root = root  # make sure all the other methods in the base class know how to find the file
                break
        return abspath


class AllowListFileHandler(tornado.web.StaticFileHandler):
    def initialize(self, allowlist=[], denylist=[], **kwargs):
        self.allowlist = allowlist
        self.denylist = denylist
        super().initialize(**kwargs)

    def get_absolute_path(self, root, path):
        # StaticFileHandler.get always calls this method first, so we use this as the
        # place to check the path. Note that now the path separator is os dependent (\\ on windows)
        allowlisted = any(re.fullmatch(pattern, path) for pattern in self.allowlist)
        denylisted = any(re.fullmatch(pattern, path) for pattern in self.denylist)
        if not allowlisted:
            raise tornado.web.HTTPError(403, "File not allowlisted")
        if denylisted:
            raise tornado.web.HTTPError(403, "File denylisted")
        return super().get_absolute_path(root, path)



# test tree rendering
import pytest


@pytest.fixture
def voila_args(notebook_directory, voila_args_extra):
    return ["--VoilaTest.root_dir=%r" % notebook_directory, *voila_args_extra]


@pytest.fixture
def jupyter_server_args_extra():
    return [
        '--VoilaConfiguration.extension_language_mapping={".xcpp": "C++11"}',
        "--VoilaConfiguration.classic_tree=true",
    ]


async def test_tree(http_server_client, base_url):
    response = await http_server_client.fetch(base_url + "voila/tree")
    assert response.code == 200
    text = response.body.decode("utf-8")
    assert "print.ipynb" in text, "tree handler should render ipynb files"
    assert (
        "print.xcpp" in text
    ), "tree handler should render xcpp files (due to extension_language_mapping)"
    assert (
        "print.py" not in text
    ), "tree handler should not render .py files (due to extension_language_mapping)"



# fixtures common for app and server
import os
import time

import pytest

BASE_DIR = os.path.dirname(__file__)


@pytest.fixture
def base_url():
    return "/"


@pytest.fixture
def notebook_directory():
    return os.path.join(BASE_DIR, "notebooks")


@pytest.fixture
def print_notebook_url(base_url):
    return base_url + "voila/render/print.ipynb"


@pytest.fixture
def syntax_error_notebook_url(base_url):
    return base_url + "voila/render/syntax_error.ipynb"


@pytest.fixture
def voila_notebook(notebook_directory):
    return os.path.join(notebook_directory, "print.ipynb")


@pytest.fixture(autouse=True)
def sleep_between_tests():
    yield
    time.sleep(1)



# tests the --template argument of Voilà
import os

import pytest


@pytest.fixture
def voila_notebook(notebook_directory):
    return os.path.join(notebook_directory, "skip-voila-cell.ipynb")


@pytest.fixture
def voila_args_extra():
    return ["--template=skip_template"]


async def test_markdown_preprocessor(http_server_client, base_url):
    response = await http_server_client.fetch(base_url)
    assert response.code == 200
    html_text = response.body.decode("utf-8")
    assert "Hi Voilà cell" in html_text
    assert "Hi non Voilà cell" not in html_text



# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.4'
#       jupytext_version: 1.2.1
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python
# ---

print("Hi Voilà!")



# tests programmatic config of template system
import os

import pytest

BASE_DIR = os.path.dirname(__file__)


@pytest.fixture
def jupyter_server_config():
    def config(app):
        pass

    os.environ["JUPYTER_CONFIG_DIR"] = os.path.join(
        BASE_DIR, "..", "configs", "general"
    )
    yield config
    del os.environ["JUPYTER_CONFIG_DIR"]


@pytest.mark.xfail(reason="needs to be fixed")
async def test_lists_extension(http_server_client, print_notebook_url):
    response = await http_server_client.fetch(print_notebook_url)
    assert response.code == 200
    html_text = response.body.decode("utf-8")
    assert "Hi Voilà" in html_text
    assert "ipytest/extension.js" in html_text



import os
from copy import deepcopy

from nbformat import NO_CONVERT, read

from voila.execute import executenb

BASE_DIR = os.path.dirname(__file__)
WIDGET_MIME_TYPE_VIEW = "application/vnd.jupyter.widget-view+json"
WIDGET_MIME_TYPE_STATE = "application/vnd.jupyter.widget-state+json"


# based on nbconvert.preprocessors.tests.test_execute.TestExecute
# we cannot import it because pytest would then also execute those tests
def normalize_output(output):
    """
    Normalizes outputs for comparison.
    """
    output = dict(output)
    if "metadata" in output:
        del output["metadata"]
    # tracebacks can be different per installation/python version
    if "traceback" in output:
        del output["traceback"]
    if "application/vnd.jupyter.widget-view+json" in output.get("data", {}):
        output["data"]["application/vnd.jupyter.widget-view+json"][
            "model_id"
        ] = "<MODEL_ID>"


def normalize_outputs(outputs):
    for output in outputs:
        normalize_output(output)


def test_execute_output():
    path = os.path.join(BASE_DIR, "notebooks/output.ipynb")
    nb = read(path, NO_CONVERT)
    nb_voila = deepcopy(nb)
    executenb(nb_voila, timeout=30)

    widget_states = nb.metadata.widgets[WIDGET_MIME_TYPE_STATE]["state"]
    widget_states_voila = nb_voila.metadata.widgets[WIDGET_MIME_TYPE_STATE]["state"]

    for cell_voila, cell in zip(nb_voila.cells, nb.cells):
        for output_voila, output in zip(cell_voila.outputs, cell.outputs):
            if "data" in output and WIDGET_MIME_TYPE_VIEW in output["data"]:
                widget_id = output["data"][WIDGET_MIME_TYPE_VIEW]["model_id"]
                widget_id_voila = output_voila["data"][WIDGET_MIME_TYPE_VIEW][
                    "model_id"
                ]
                widget_state = widget_states[widget_id]
                widget_state_voila = widget_states_voila[widget_id_voila]
                # if the widget is an output widget, it has the outputs, which we also check
                assert normalize_outputs(
                    widget_state.state.get("outputs", [])
                ) == normalize_outputs(widget_state_voila.state.get("outputs", []))
            normalize_output(output_voila)
            normalize_output(output)
            assert output_voila == output



# tests the --template argument of voila
import base64
import os

import pytest

NOTEBOOK_PATH = "images.ipynb"


@pytest.fixture
def voila_notebook(notebook_directory):
    return os.path.join(notebook_directory, NOTEBOOK_PATH)


async def test_image_inlining(http_server_client, base_url, notebook_directory):
    response = await http_server_client.fetch(base_url)
    html_text = response.body.decode("utf-8")

    assert "data:image/svg+xml;base64," in html_text
    assert "data:image/png;base64," in html_text

    # check if the external file is inline
    with open(
        os.path.join(notebook_directory, notebook_directory, "jupyter.svg"), "rb"
    ) as f:
        svg_data = f.read()
    svg_data_base64 = base64.b64encode(svg_data).decode("ascii")
    assert svg_data_base64 in html_text



import os

from setuptools import setup

data_files = []
for dirpath, _dirnames, filenames in os.walk("share/jupyter/voila/templates"):
    if filenames:
        data_files.append(
            (dirpath, [os.path.join(dirpath, filename) for filename in filenames])
        )


setup(
    name="skip_template",
    version="0.0.1",
    description="Test template for Voilà",
    data_files=data_files,
    include_package_data=True,
    author="Voilà Development team",
    author_email="jupyter@googlegroups.com",
)



import asyncio
import os
import time

import pytest

BASE_DIR = os.path.dirname(__file__)
NOTEBOOK_EXECUTION_TIME = 3
NUMBER_PREHEATED_KERNEL = 2
TIME_THRESHOLD = 1


@pytest.fixture
def voila_config_file_paths_arg():
    path = os.path.join(BASE_DIR, "..", "configs", "preheat")
    return "--VoilaTest.config_file_paths=[%r]" % path


@pytest.fixture
def preheat_mode():
    return True


@pytest.fixture
def voila_notebook(notebook_directory):
    return os.path.join(notebook_directory, "preheat", "pre_heat.ipynb")


async def send_request(sc, url, wait=0):
    await asyncio.sleep(wait)
    real_time = time.time()
    response = await sc.fetch(url)
    real_time = time.time() - real_time
    html_text = response.body.decode("utf-8")
    return real_time, html_text


async def test_refill_kernel_asynchronously(http_server_client, base_url):
    await asyncio.sleep(NUMBER_PREHEATED_KERNEL * NOTEBOOK_EXECUTION_TIME + 1)
    fast = []
    slow = []
    for _i in range(5 * NUMBER_PREHEATED_KERNEL):
        time, _ = await send_request(sc=http_server_client, url=base_url)
        if time < TIME_THRESHOLD:
            fast.append(time)
        else:
            slow.append(time)

    assert len(fast) > 1
    assert len(slow) > 1
    assert len(fast) + len(slow) == 5 * NUMBER_PREHEATED_KERNEL
    await asyncio.sleep(NOTEBOOK_EXECUTION_TIME + 1)


async def test_env_variable_defined_in_kernel(http_server_client, base_url):
    await asyncio.sleep(NUMBER_PREHEATED_KERNEL * NOTEBOOK_EXECUTION_TIME + 1)
    _, text = await send_request(sc=http_server_client, url=base_url)
    assert "bar" in text
    await asyncio.sleep(NOTEBOOK_EXECUTION_TIME + 1)



import re
import pytest


@pytest.fixture
def voila_config():
    def config(voila_app):
        voila_app.tornado_settings["disable_check_xsrf"] = True

    return config


async def test_shutdown_handler(http_server_client, base_url):
    response = await http_server_client.fetch(base_url)
    html_text = response.body.decode("utf-8")
    pattern = r"""kernelId": ["']([0-9a-zA-Z-]+)["']"""
    groups = re.findall(pattern, html_text)
    kernel_id = groups[0]
    shutdown_url = f"{base_url}voila/api/shutdown/{kernel_id}"
    shutdown_response = await http_server_client.fetch(
        shutdown_url, method="POST", body=b""
    )
    assert shutdown_response.code == 204



import json
import pytest
import tornado


@pytest.fixture
def preheat_mode():
    return False


@pytest.fixture
def contents_prefix(base_url):
    return base_url + "voila/api/contents"


@pytest.fixture
def voila_args(notebook_directory, voila_args_extra):
    return ["--VoilaTest.root_dir=%r" % notebook_directory, *voila_args_extra]


@pytest.fixture
def voila_args_extra():
    return ['--VoilaConfiguration.extension_language_mapping={".py": "python"}']


async def test_contents_endpoint(http_server_client, contents_prefix):
    response = await http_server_client.fetch(contents_prefix)
    html_json = json.loads(response.body.decode("utf-8"))
    assert "content" in html_json
    assert len(html_json["content"]) > 0


async def test_get_notebook(http_server_client, contents_prefix):
    response = await http_server_client.fetch(contents_prefix + "/autokill.ipynb")
    html_json = json.loads(response.body.decode("utf-8"))
    assert html_json["name"] == "autokill.ipynb"
    assert html_json["content"] is None


async def test_get_not_allowed_file(http_server_client, contents_prefix):
    with pytest.raises(tornado.httpclient.HTTPClientError):
        await http_server_client.fetch(contents_prefix + "/print.xcpp")


async def test_get_allowed_file(http_server_client, contents_prefix):
    response = await http_server_client.fetch(contents_prefix + "/print.py")
    html_json = json.loads(response.body.decode("utf-8"))
    assert html_json["name"] == "print.py"
    assert html_json["content"] is None



import os

import pytest

BASE_DIR = os.path.dirname(__file__)


@pytest.fixture
def voila_args_extra():
    return ["--VoilaConfiguration.strip_sources=False", "--VoilaExecutor.timeout=240"]


async def test_no_strip_sources(http_server_client, base_url):
    response = await http_server_client.fetch(base_url)
    assert response.code == 200
    html_text = response.body.decode("utf-8")
    assert "Hi Voilà" in html_text
    assert "print" in html_text, "the source code should *NOT* be stripped"



# tests prelaunch hook config
import os
from urllib.parse import quote_plus

import pytest

BASE_DIR = os.path.dirname(__file__)


@pytest.fixture
def voila_notebook(notebook_directory):
    return os.path.join(notebook_directory, "print_parameterized.ipynb")


@pytest.fixture
def voila_config():
    def parameterize_with_papermill(req, notebook, cwd):
        import tornado

        # Grab parameters
        parameters = req.get_argument("parameters", {})

        # try to convert to dict if not e.g. string/unicode
        if not isinstance(parameters, dict):
            try:
                parameters = tornado.escape.json_decode(parameters)
            except ValueError:
                parameters = None

        # if passed and a dict, use papermill to inject parameters
        if parameters and isinstance(parameters, dict):
            from papermill.parameterize import parameterize_notebook

            # setup for papermill
            #
            # these two blocks are done
            # to avoid triggering errors
            # in papermill's notebook
            # loading logic
            for cell in notebook.cells:
                if "tags" not in cell.metadata:
                    cell.metadata.tags = []
                if "papermill" not in notebook.metadata:
                    notebook.metadata.papermill = {}

            # Parameterize with papermill
            return parameterize_notebook(notebook, parameters)

    def config(app):
        app.prelaunch_hook = parameterize_with_papermill

    return config


async def test_prelaunch_hook_papermill(http_server_client, base_url):
    url = base_url + "?parameters=" + quote_plus('{"name":"Parameterized_Variable"}')
    response = await http_server_client.fetch(url)
    assert response.code == 200
    html_text = response.body.decode("utf-8")
    assert "Hi Parameterized_Variable" in html_text
    assert (
        "test_template.css" not in html_text
    ), "test_template should not be the default"



import pytest

NOTEBOOK_PATH = "cgi.ipynb"


@pytest.fixture
def notebook_cgi_path(base_url):
    return base_url + f"voila/render/{NOTEBOOK_PATH}"


@pytest.fixture
def voila_args(notebook_directory, voila_args_extra):
    return ["--VoilaTest.root_dir=%r" % notebook_directory, *voila_args_extra]


async def test_cgi_using_query_parameters(http_server_client, notebook_cgi_path):
    response = await http_server_client.fetch(notebook_cgi_path + "?username=VOILA")
    assert response.code == 200
    assert "VOILA" in response.body.decode("utf-8")



# NOTE: this is a duplicate of ../app/cwd_subdir_test.py
# we might want to find a better pattern of executing test for the app and extension
import pytest


@pytest.fixture
def cwd_subdir_notebook_url(base_url):
    return base_url + "voila/render/subdir/cwd_subdir.ipynb"


@pytest.fixture
def voila_args(notebook_directory, voila_args_extra):
    return ["--VoilaTest.root_dir=%r" % notebook_directory, *voila_args_extra]


async def test_hello_world(http_server_client, cwd_subdir_notebook_url):
    response = await http_server_client.fetch(cwd_subdir_notebook_url)
    html_text = response.body.decode("utf-8")
    assert "check for the cwd" in html_text



import os

import pytest


@pytest.fixture
def voila_notebook(notebook_directory):
    return os.path.join(notebook_directory, "sleep.ipynb")


@pytest.fixture
def voila_args_extra():
    return ["--VoilaExecutor.timeout=1", "--KernelManager.shutdown_wait_time=0.1"]


async def test_timeout(http_server_client, base_url):
    response = await http_server_client.fetch(base_url)
    html_text = response.body.decode("utf-8")
    assert "Cell execution timed out" in html_text



# simply tests if some notebooks execute
import pytest

NOTEBOOK_PATH = "other_comms.ipynb"


@pytest.fixture
def notebook_other_comms_path(base_url):
    return base_url + f"voila/render/{NOTEBOOK_PATH}"


@pytest.fixture
def voila_args(notebook_directory, voila_args_extra):
    return ["--VoilaTest.root_dir=%r" % notebook_directory, *voila_args_extra]


async def test_other_comms(http_server_client, notebook_other_comms_path):
    response = await http_server_client.fetch(notebook_other_comms_path)
    html_text = response.body.decode("utf-8")
    assert "This notebook executed" in html_text



# tests prelaunch hook config
import os

import pytest
from nbformat import NotebookNode

BASE_DIR = os.path.dirname(__file__)


@pytest.fixture
def voila_notebook(notebook_directory):
    return os.path.join(notebook_directory, "print.ipynb")


@pytest.fixture
def voila_config():
    def foo(req, notebook, cwd):
        argument = req.get_argument("test")
        notebook.cells.append(
            NotebookNode(
                {
                    "cell_type": "code",
                    "execution_count": 0,
                    "metadata": {},
                    "outputs": [],
                    "source": f'print("Hi prelaunch hook {argument}!")\n',
                }
            )
        )

    def config(app):
        app.prelaunch_hook = foo

    return config


async def test_prelaunch_hook(http_server_client, base_url):
    response = await http_server_client.fetch(
        base_url + "?test=blerg",
    )
    assert response.code == 200
    assert "Hi Voilà" in response.body.decode("utf-8")
    assert "Hi prelaunch hook blerg" in response.body.decode("utf-8")



import pytest
import tornado

from voila.static_file_handler import TemplateStaticFileHandler


async def test_static_file_absolute_path(app, base_url, http_server_client):
    response_lab = await http_server_client.fetch(
        f"{base_url}voila/templates/lab/static/voila.js"
    )
    assert response_lab.code == 200
    abspath = TemplateStaticFileHandler.get_absolute_path(None, "lab/static/voila.js")
    with open(abspath, encoding="utf-8", errors="replace") as f:
        content = f.read()
    assert response_lab.body.decode("utf-8") == content


async def test_static_file_not_found(app, base_url, http_server_client):
    with pytest.raises(tornado.httpclient.HTTPClientError, match="HTTP 404.*"):
        await http_server_client.fetch(
            f"{base_url}voila/templates/lab/static/doesnotexist.js"
        )


async def test_static_file_availability_default(app, base_url, http_server_client):
    response_lab = await http_server_client.fetch(
        f"{base_url}voila/templates/lab/static/voila.js"
    )
    response_default = await http_server_client.fetch(
        f"{base_url}voila/static/voila.js"
    )
    assert response_lab.code == 200
    assert response_default.code == 200
    assert response_lab.body.decode("utf-8") == response_default.body.decode("utf-8")


async def test_static_file_override(app, base_url, http_server_client):
    response_lab = await http_server_client.fetch(
        f"{base_url}voila/templates/lab/static/voila.js"
    )
    response_test_template = await http_server_client.fetch(
        f"{base_url}voila/templates/test_template/static/voila.js"
    )
    assert response_lab.code == 200
    assert response_test_template.code == 200
    assert response_lab.body.decode("utf-8") != response_test_template.body.decode(
        "utf-8"
    )


async def test_static_file_other_template(app, base_url, http_server_client):
    response = await http_server_client.fetch(
        f"{base_url}voila/templates/test_template/static/only-in-test-template.js"
    )
    assert response.code == 200
    assert response.body.decode("utf-8") == "", "empty file expected"



import pytest

NOTEBOOK_PATH = "syntax_error.ipynb"


@pytest.fixture(params=[True, False])
def show_tracebacks(request):
    return request.param


@pytest.fixture
def notebook_show_traceback_path(base_url):
    return base_url + f"voila/render/{NOTEBOOK_PATH}"


@pytest.fixture
def voila_args(notebook_directory, voila_args_extra, show_tracebacks):
    return [
        "--VoilaTest.root_dir=%r" % notebook_directory,
        f"--VoilaConfiguration.show_tracebacks={show_tracebacks}",
        *voila_args_extra,
    ]


async def test_syntax_error(
    http_server_client, notebook_show_traceback_path, show_tracebacks
):
    response = await http_server_client.fetch(notebook_show_traceback_path)
    assert response.code == 200
    output = response.body.decode("utf-8")
    if show_tracebacks:
        assert "this is a syntax error" in output, 'should show the "code"'
        assert (
            "SyntaxError" in output and "invalid syntax" in output
        ), "should show the error"
    else:
        assert "There was an error when executing cell" in output
        assert "This should not be executed" not in output



import asyncio
import os
import time

import pytest


@pytest.fixture
def preheat_mode():
    return True


@pytest.fixture
def voila_notebook(notebook_directory):
    return os.path.join(notebook_directory, "preheat", "pre_heat.ipynb")


NOTEBOOK_EXECUTION_TIME = 2
TIME_THRESHOLD = NOTEBOOK_EXECUTION_TIME


async def send_request(sc, url, wait=0):
    await asyncio.sleep(wait)
    real_time = time.time()
    response = await sc.fetch(url)
    real_time = time.time() - real_time
    html_text = response.body.decode("utf-8")
    return real_time, html_text


async def test_request_before_kernel_heated(http_server_client, base_url):
    """
    We send a request to server immediately, a normal kernel
    will be used so it should take more than `NOTEBOOK_EXECUTION_TIME`
    to display result.
    """
    time, text = await send_request(sc=http_server_client, url=base_url)
    assert "hello world" in text
    assert time > NOTEBOOK_EXECUTION_TIME
    await asyncio.sleep(NOTEBOOK_EXECUTION_TIME + 1)


async def test_render_time_with_preheated_kernel(http_server_client, base_url):
    """
    We wait for kernel properly heated before sending request,
    it should take under `TIME_THRESHOLD` second to return result
    """
    time, text = await send_request(
        sc=http_server_client, url=base_url, wait=NOTEBOOK_EXECUTION_TIME + 2
    )
    assert "hello world" in text
    assert time < TIME_THRESHOLD
    await asyncio.sleep(NOTEBOOK_EXECUTION_TIME + 1)


async def test_render_time_with_multiple_requests(http_server_client, base_url):
    """
    We send a second request just after the first one, so the
    pool is not filled yet and a normal kernel is used instead
    """
    time_list = []
    for wait in [NOTEBOOK_EXECUTION_TIME + 1, 0]:
        time, _ = await send_request(sc=http_server_client, url=base_url, wait=wait)
        time_list.append(time)

    assert (
        time_list[1] > time_list[0]
    )  # Render time for a normal kernel is bigger than a a preheated kernel
    await asyncio.sleep(NOTEBOOK_EXECUTION_TIME + 1)


async def test_request_with_query(http_server_client, base_url):
    """
    We sent request with query parameter, preheat kernel should
    be activated.
    """
    url = f"{base_url}?foo=bar"
    time, _ = await send_request(
        sc=http_server_client, url=url, wait=NOTEBOOK_EXECUTION_TIME + 1
    )
    assert time < TIME_THRESHOLD
    await asyncio.sleep(NOTEBOOK_EXECUTION_TIME + 1)


async def test_request_with_theme_parameter(http_server_client, base_url):
    """
    We sent request with theme parameter, preheat kernel should
    be used if requested theme is same as theme used for prerendered.
    """
    wait = NOTEBOOK_EXECUTION_TIME + 2

    url = f"{base_url}?theme=dark"
    time, _ = await send_request(sc=http_server_client, url=url, wait=wait)
    assert time > TIME_THRESHOLD

    url = f"{base_url}?theme=light"
    time, _ = await send_request(sc=http_server_client, url=url, wait=wait)
    assert time < TIME_THRESHOLD


async def test_request_with_template_parameter(http_server_client, base_url):
    """
    We sent request with theme parameter, preheat kernel should
    be used if requested theme is same as theme used for prerendered.
    """
    wait = NOTEBOOK_EXECUTION_TIME + 2

    url = f"{base_url}?template=lab"
    time, _ = await send_request(sc=http_server_client, url=url, wait=wait)
    assert time < TIME_THRESHOLD



import asyncio
import os
import time

import pytest


@pytest.fixture(params=[['--Voila.base_url="/base/"'], []])
def using_base_url(request):
    return request.param


@pytest.fixture(params=[['--Voila.server_url="/server/"'], []])
def using_server_url(request):
    return request.param


@pytest.fixture()
def voila_args_extra(http_server_port, using_server_url, using_base_url):
    return [f"--port={http_server_port[-1]}", *using_server_url, *using_base_url]


@pytest.fixture
def preheat_mode():
    return True


@pytest.fixture
def voila_notebook(notebook_directory):
    return os.path.join(notebook_directory, "preheat", "get_query_string.ipynb")


NOTEBOOK_EXECUTION_TIME = 2
TIME_THRESHOLD = NOTEBOOK_EXECUTION_TIME


async def send_request(sc, url, wait=0):
    await asyncio.sleep(wait)
    real_time = time.time()
    response = await sc.fetch(url)
    real_time = time.time() - real_time
    html_text = response.body.decode("utf-8")
    return real_time, html_text


async def test_request_with_query(
    http_server_client, base_url, using_base_url, using_server_url
):
    """
    We sent request with query parameter, `get_query_string` should
    return value.
    """
    if len(using_server_url) > 0:
        url = "/server/?foo=bar"
    else:
        if len(using_base_url) > 0:
            url = "/base/?foo=bar"
        else:
            url = f"{base_url}?foo=bar"
    _, html_text = await send_request(
        sc=http_server_client, url=url, wait=NOTEBOOK_EXECUTION_TIME + 1
    )
    assert "foo=bar" in html_text



import os

import pytest

TEST_XEUS_CLING = os.environ.get("VOILA_TEST_XEUS_CLING", "") == "1"
NOTEBOOK_PATH = "print.xcpp"


@pytest.fixture
def cpp_file_url(base_url, preheat_mode):
    return base_url + f"voila/render/{NOTEBOOK_PATH}"


@pytest.fixture
def voila_args_extra():
    return [
        '--VoilaConfiguration.extension_language_mapping={".xcpp": "C++11"}',
        "--VoilaExecutor.timeout=240",
    ]


@pytest.fixture
def voila_args(notebook_directory, voila_args_extra, preheat_mode):
    return ["--VoilaTest.root_dir=%r" % notebook_directory, *voila_args_extra]


@pytest.mark.skipif(
    not TEST_XEUS_CLING, reason="opt in to avoid having to install xeus-cling"
)
async def test_non_existing_kernel(http_server_client, cpp_file_url):
    response = await http_server_client.fetch(cpp_file_url)
    assert response.code == 200
    assert "Hello Voilà, from c++" in response.body.decode("utf-8")



# test serving a notebook or python/c++ notebook
import os

import pytest

TEST_XEUS_CLING = os.environ.get("VOILA_TEST_XEUS_CLING", "") == "1"


@pytest.fixture
def preheat_mode():
    return False


@pytest.fixture
def voila_args(notebook_directory, voila_args_extra):
    return ["--VoilaTest.root_dir=%r" % notebook_directory, *voila_args_extra]


async def test_print(http_server_client, print_notebook_url):
    print(print_notebook_url)
    response = await http_server_client.fetch(print_notebook_url)
    assert response.code == 200
    assert "Hi Voilà" in response.body.decode("utf-8")


@pytest.fixture
def voila_args_extra():
    return ['--VoilaConfiguration.extension_language_mapping={".py": "python"}']


async def test_print_py(http_server_client, print_notebook_url):
    print(print_notebook_url)
    response = await http_server_client.fetch(print_notebook_url.replace("ipynb", "py"))
    assert response.code == 200
    assert "Hi Voilà" in response.body.decode("utf-8")


@pytest.mark.skipif(
    not TEST_XEUS_CLING, reason="opt in to avoid having to install xeus-cling"
)
async def test_print_julia_notebook(http_server_client, print_notebook_url):
    print(print_notebook_url)
    response = await http_server_client.fetch(
        print_notebook_url.replace(".ipynb", "_cpp.ipynb")
    )
    assert response.code == 200
    assert "Hi Voilà, from c++" in response.body.decode("utf-8")



# tests programmatic config of template system
import os

import pytest

BASE_DIR = os.path.dirname(__file__)


@pytest.fixture
def voila_args_extra():
    path_test_template = os.path.abspath(
        os.path.join(
            BASE_DIR, "../test_template/share/jupyter/voila/templates/test_template/"
        )
    )
    path_default = os.path.abspath(
        os.path.join(BASE_DIR, "../../share/jupyter/voila/templates/default")
    )
    return [
        "--template=test_template",
        "--VoilaTest.template_paths=[{!r}, {!r}]".format(
            path_test_template, path_default
        ),
        "--VoilaExecutor.timeout=240",
    ]


async def test_template_test(http_server_client, base_url):
    response = await http_server_client.fetch(base_url)
    assert response.code == 200
    assert "test_template.css" in response.body.decode("utf-8")



# test basics of voila running a notebook
import asyncio
import json
import re

import tornado.web

try:
    from unittest import mock
except ImportError:
    from unittest import mock


async def test_hello_world(http_server_client, base_url):
    response = await http_server_client.fetch(base_url)
    assert response.code == 200
    html_text = response.body.decode("utf-8")
    assert "Hi Voilà" in html_text
    assert "print(" not in html_text, "by default the source code should be stripped"
    assert (
        "test_template.css" not in html_text
    ), "test_template should not be the default"


async def test_no_execute_allowed(
    voila_app, app, http_server_client, base_url, http_server_port
):
    assert voila_app.app is app
    response = (await http_server_client.fetch(base_url)).body.decode("utf-8")
    pattern = r"""kernelId": ["']([0-9a-zA-Z-]+)["']"""
    groups = re.findall(pattern, response)
    kernel_id = groups[0]
    print(kernel_id, base_url)
    session_id = "445edd75-c6f5-45d2-8b58-5fe8f84a7123"
    url = f"ws://localhost:{http_server_port[1]}{base_url}api/kernels/{kernel_id}/channels?session_id={session_id}"
    conn = await tornado.websocket.websocket_connect(url)

    msg = {
        "header": {
            "msg_id": "8573fb401ac848aab63c3bf0081e9b65",
            "username": "username",
            "session": "7a7d94334ea745f888d9c479fa738d63",
            "msg_type": "execute_request",
            "version": "5.2",
        },
        "metadata": {},
        "content": {
            "code": "print('la')",
            "silent": False,
            "store_history": False,
            "user_expressions": {},
            "allow_stdin": False,
            "stop_on_error": False,
        },
        "buffers": [],
        "parent_header": {},
        "channel": "shell",
    }
    with mock.patch.object(voila_app.log, "warning") as mock_warning:
        await conn.write_message(json.dumps(msg))
        # make sure the warning method is called
        while not mock_warning.called:
            await asyncio.sleep(0.1)
    mock_warning.assert_called_with(
        'Received message of type "execute_request", which is not allowed. Ignoring.'
    )



# tests the --template argument of Voilà
import pytest


@pytest.fixture
def voila_args_extra():
    return ["--template=test_template", "--VoilaExecutor.timeout=240"]


async def test_template(http_server_client, base_url):
    response = await http_server_client.fetch(base_url)
    assert response.code == 200
    assert "test_template.css" in response.body.decode("utf-8")
    assert "Hi Voilà" in response.body.decode("utf-8")



import pytest

NOTEBOOK_PATH = "no_kernelspec.ipynb"


@pytest.fixture
def non_existing_kernel_notebook(base_url):
    return base_url + f"voila/render/{NOTEBOOK_PATH}"


@pytest.fixture
def voila_args(notebook_directory, voila_args_extra):
    return ["--VoilaTest.root_dir=%r" % notebook_directory, *voila_args_extra]


async def test_non_existing_kernel(
    http_server_client,
    non_existing_kernel_notebook,
):
    response = await http_server_client.fetch(non_existing_kernel_notebook)
    assert response.code == 200
    assert "Executing without a kernelspec" in response.body.decode("utf-8")



# tests the --template argument of voila
import os

import pytest


@pytest.fixture
def voila_notebook(notebook_directory):
    return os.path.join(notebook_directory, "cwd.ipynb")


async def test_template_cwd(http_server_client, base_url):
    response = await http_server_client.fetch(base_url)
    html_text = response.body.decode("utf-8")
    assert "check for the cwd" in html_text



"""Tests loading template of jinja2 templates"""
import os

from jinja2 import Environment, FileSystemLoader

from voila.paths import collect_paths

HERE = os.path.dirname(__file__)

ROOT_DIRS = [os.path.join(HERE, "user"), os.path.join(HERE, "system")]


def test_loader_default_nbconvert():
    paths = collect_paths(["nbconvert"], "default", root_dirs=ROOT_DIRS)
    loader = FileSystemLoader(paths)
    env = Environment(loader=loader)
    template = env.get_template("index.tpl")
    output = template.render()
    assert "this is block base:nested in nbconvert/default/index.tpl" in output


def test_loader_foo():
    paths = collect_paths(["voila", "nbconvert"], "foo", root_dirs=ROOT_DIRS)
    loader = FileSystemLoader(paths)
    env = Environment(loader=loader)
    template = env.get_template("index.tpl")
    output = template.render()
    assert "this is block base:nested in voila/default/index.tpl" in output
    assert "this is block base:nested in voila/foo/index.tpl" in output
    assert "this is block base:nested in nbconvert/foo/index.tpl" in output
    assert "this is block base:nested in nbconvert/default/index.tpl" not in output


def test_loader_bar_voila():
    paths = collect_paths(["voila", "nbconvert"], "bar", root_dirs=ROOT_DIRS)
    loader = FileSystemLoader(paths)
    env = Environment(loader=loader)
    template = env.get_template("index.tpl")
    output = template.render()
    assert "this is block base in nbconvert/bar/index.tpl" in output
    assert "this is block base in nbconvert/default/index.tpl" in output
    assert "this is block base:nested in voila/default/index.tpl" in output
    assert "this is block base:nested2 in nbconvert/default/index.tpl" in output
    assert "this is block common in nbconvert/bar/parent.tpl" in output


def test_loader_bar_nbconvert():
    paths = collect_paths(["nbconvert"], "bar", root_dirs=ROOT_DIRS)
    loader = FileSystemLoader(paths)
    env = Environment(loader=loader)
    template = env.get_template("index.tpl")
    output = template.render()
    assert "this is block base in nbconvert/bar/index.tpl" in output
    assert "this is block base in nbconvert/default/index.tpl" in output
    assert "this is block base:nested in nbconvert/default/index.tpl" in output
    assert "this is block base:nested2 in nbconvert/default/index.tpl" in output
    assert "this is block common in nbconvert/bar/parent.tpl" in output



#############################################################################
# Copyright (c) 2018, Voilà Contributors                                    #
# Copyright (c) 2018, QuantStack                                            #
#                                                                           #
# Distributed under the terms of the BSD 3-Clause License.                  #
#                                                                           #
# The full license is in the file LICENSE, distributed with this software.  #
#############################################################################
import os

from jupyter_server.utils import url_escape, url_path_join
from jupyter_core.utils import ensure_async
from tornado import web

from ..treehandler import VoilaTreeHandler
from ..utils import (
    create_include_assets_functions,
    get_page_config,
    get_server_root_dir,
)


class TornadoVoilaTreeHandler(VoilaTreeHandler):
    @web.authenticated
    async def get(self, path=""):
        cm = self.contents_manager
        dir_exists = await ensure_async(cm.dir_exists(path=path))
        file_exists = await ensure_async(cm.file_exists(path))
        if dir_exists:
            is_hidden = await ensure_async(cm.is_hidden(path))
            if is_hidden and not cm.allow_hidden:
                self.log.info("Refusing to serve hidden directory, via 404 Error")
                raise web.HTTPError(404)
            breadcrumbs = self.generate_breadcrumbs(path)
            page_title = self.generate_page_title(path)
            contents = await ensure_async(cm.get(path))

            def allowed_content(content):
                if content["type"] in ["directory", "notebook"]:
                    return True
                __, ext = os.path.splitext(content.get("path"))
                return ext in self.allowed_extensions

            contents["content"] = sorted(contents["content"], key=lambda i: i["name"])
            contents["content"] = filter(allowed_content, contents["content"])
            theme_arg = (
                self.get_argument("theme", self.voila_configuration.theme)
                if self.voila_configuration.allow_theme_override == "YES"
                else self.voila_configuration.theme
            )
            classic_tree_arg = self.get_argument("classic-tree", "").lower()
            if classic_tree_arg == "true":
                classic_tree = True
            elif classic_tree_arg == "false":
                classic_tree = False
            else:
                classic_tree = self.voila_configuration.classic_tree

            theme_arg = self.validate_theme(theme_arg, classic_tree)

            page_config = get_page_config(
                base_url=self.base_url,
                settings=self.settings,
                log=self.log,
                voila_configuration=self.voila_configuration,
            )
            page_config["jupyterLabTheme"] = theme_arg
            page_config["frontend"] = "voila"
            page_config["query"] = self.request.query
            template_name = "tree-lab.html" if not classic_tree else "tree.html"

            resources = create_include_assets_functions(template_name, self.base_url)
            self.write(
                self.render_template(
                    template_name,
                    frontend="voila",
                    main_js="voila.js",
                    page_title=page_title,
                    notebook_path=path,
                    breadcrumbs=breadcrumbs,
                    contents=contents,
                    terminals_available=False,
                    server_root=get_server_root_dir(self.settings),
                    query=self.request.query,
                    page_config=page_config,
                    theme=theme_arg,
                    resources=resources,
                )
            )
        elif file_exists:
            # it's not a directory, we have redirecting to do
            model = await ensure_async(cm.get(path, content=False))
            # redirect to /api/notebooks if it's a notebook, otherwise /api/files
            service = "notebooks" if model["type"] == "notebook" else "files"
            url = url_path_join(
                self.base_url,
                service,
                url_escape(path),
            )
            self.log.debug("Redirecting %s to %s", self.request.path, url)
            self.redirect(url)
        else:
            raise web.HTTPError(404)



# tests basic things the templates should have
import os

import pytest

BASE_DIR = os.path.dirname(__file__)


@pytest.fixture(params=["lab", "classic"])
def voila_args_extra(request):
    return [f"--template={request.param}"]


async def test_lists_extension(http_server_client, base_url, voila_app):
    response = await http_server_client.fetch(base_url)
    assert response.code == 200
    html_text = response.body.decode("utf-8")
    assert "/voila/static/voila.js" in html_text



import os

import pytest

MAX_TIMEOUT_SECONDS = 240


@pytest.fixture
def preheat_mode():
    return False


@pytest.fixture
def voila_args_extra():
    return [f"--VoilaExecutor.timeout={MAX_TIMEOUT_SECONDS}"]


@pytest.fixture
def voila_notebook(notebook_directory):
    return os.path.join(notebook_directory, "many_iopub_messages.ipynb")


@pytest.mark.filterwarnings("ignore")
async def test_template_cwd(http_server_client, base_url):
    response = await http_server_client.fetch(
        base_url, request_timeout=MAX_TIMEOUT_SECONDS + 20
    )
    html_text = response.body.decode("utf-8")
    assert "you should see me" in html_text



#############################################################################
# Copyright (c) 2021, Voilà Contributors                                    #
# Copyright (c) 2021, QuantStack                                            #
#                                                                           #
# Distributed under the terms of the BSD 3-Clause License.                  #
#                                                                           #
# The full license is in the file LICENSE, distributed with this software.  #
#############################################################################


import asyncio
import os
import re
from pathlib import Path
from typing import Awaitable
from typing import Dict as TypeDict
from typing import List as TypeList
from typing import Tuple, Type, TypeVar, Union
from warnings import warn

from jupyter_core.utils import ensure_async
from traitlets import validate
from traitlets.traitlets import Dict, Float, List, default

from .notebook_renderer import NotebookRenderer
from .utils import ENV_VARIABLE, get_page_config

T = TypeVar("T")


async def wait_before(delay: float, aw: Awaitable) -> Awaitable:
    await asyncio.sleep(delay)
    return await aw


def voila_kernel_manager_factory(
    base_class: Type[T], preheat_kernel: bool, default_pool_size: int
) -> T:
    """
    Decorator used to make a normal kernel manager compatible with pre-heated
    kernel system.
    - If `preheat_kernel` is `False`, only the property
    `notebook_data` and the method `get_pool_size` are added to keep `VoilaHandler`
    working, the kernel manager will work as it is.
    - If `preheat_kernel` is `True`, the input class is transformed to
     `VoilaKernelManager` with all the functionalities.

    Args:
        - base_class (Type[T]): The kernel manager class
        - preheat_kernel (Bool): Flag to decorate the input class
        - default_pool_size (int): Size of pre-heated kernel pool for each notebook.
            Zero or negative number means disabled

    Returns:
        T: Decorated class
    """
    if not preheat_kernel:

        class NormalKernelManager(base_class):
            @property
            def notebook_data(self) -> TypeDict:
                return {}

            def get_pool_size(self, nb: str) -> int:
                return 0

        return NormalKernelManager

    else:

        class VoilaKernelManager(base_class):
            """This class adds pooling heated kernels and pre-rendered notebook
            feature to a normal kernel manager. The 'pooling heated kernels'
            part is heavily inspired from `hotpot_km`(https://github.com/voila-dashboards/hotpot_km) library.
            """

            kernel_pools_config = Dict(
                config=True,
                help="""Mapping from notebook name to the kernel configuration
                like: number of started kernels to keep on standby, environment
                variables used to start kernel""",
            )

            default_env_variables = Dict(
                {},
                config=True,
                help="""Default environmental variables for kernels
                """,
            )

            preheat_denylist = List(
                [],
                config=True,
                help="List of notebooks which do not use pre-heated kernel.",
            )

            preheat_blacklist = List(
                [],
                config=True,
                help="Deprecated, use `preheat_denylist`.",
            )

            @validate("preheat_blacklist")
            def _valid_preheat_blacklist(self, proposal):
                warn(
                    "Deprecated, use preheat_denylist instead.",
                    DeprecationWarning,
                    stacklevel=2,
                )
                return proposal["value"]

            fill_delay = Float(
                1,
                config=True,
                help="Wait time before re-filling the pool after a kernel is used",
            )

            @default("kernel_pools_config")
            def _kernel_pools_config(self):
                return {
                    "default": {
                        "pool_size": max(default_pool_size, 0),
                        "kernel_env_variables": self.default_env_variables,
                    }
                }

            def __init__(self, **kwargs):
                super().__init__(**kwargs)
                self._wait_at_startup = True
                self.notebook_data: TypeDict = {}
                self._pools: TypeDict[str, List[TypeDict]] = {}
                self.root_dir = self.parent.root_dir
                if self.parent.notebook_path is not None:
                    self.notebook_path = os.path.relpath(
                        self.parent.notebook_path, self.root_dir
                    )
                    self.fill_if_needed(delay=0, notebook_name=self.notebook_path)
                else:
                    self.notebook_path = None
                    all_notebooks = [
                        x.relative_to(self.root_dir)
                        for x in list(Path(self.root_dir).rglob("*.ipynb"))
                        if self._notebook_filter(x)
                    ]
                    for nb in all_notebooks:
                        self.fill_if_needed(delay=0, notebook_name=str(nb))

            async def get_rendered_notebook(
                self,
                notebook_name: str,
                extra_kernel_env_variables: dict = {},
                **kwargs,
            ) -> Tuple[asyncio.Task, TypeList[str]]:
                """Get the notebook rendering task and the rendered cell.
                By setting the `stop_generator` to True, the running task
                `render_task` used for rendering notebook will be stopped
                after finishing the running cell. The results of this task
                will contain the rendered cells and a generator for continuing
                render the remaining cells. We need to return also
                `renderer.rendered_cache` since it contains the rendered cells
                before the moment we set `stop_generator` to `True`, so that
                we can flush data immediately without waiting for running cell
                to be finished.

                Args:
                    notebook_name (str): Path to notebook

                Raises:
                    NameError: Raised if no notebook is provided.
                    Exception: Raised if the kernel pool is empty.

                Returns:
                    Tuple[asyncio.Task, List[str]]:

                """
                if notebook_name is None:
                    raise NameError("Notebook name must be provided!")

                if len(self._pools.get(notebook_name, ())) == 0:
                    raise Exception(f"Kernel pool for {notebook_name} is empty!")

                pool_item = self._pools[notebook_name].pop(0)
                content = await pool_item
                renderer: NotebookRenderer = content["renderer"]
                render_task: asyncio.Task = content["task"]
                kernel_id: str = content["kernel_id"]
                renderer.stop_generator = True
                self.log.info(
                    "Using pre-heated kernel: %s for %s", kernel_id, notebook_name
                )
                self.fill_if_needed(
                    delay=None,
                    notebook_name=notebook_name,
                    extra_kernel_env_variables=extra_kernel_env_variables,
                    **kwargs,
                )

                return render_task, renderer.rendered_cache, kernel_id

            def get_pool_size(self, notebook_name: str) -> int:
                return len(self._pools.get(notebook_name, []))

            def fill_if_needed(
                self,
                delay: Union[float, None] = None,
                notebook_name: Union[str, None] = None,
                extra_kernel_env_variables: dict = {},
                **kwargs,
            ) -> None:
                """Start kernels until the pool is full

                Args:
                    - delay (Union[float, None], optional): Delay time before
                    starting refill kernel. Defaults to None.
                    - notebook_name (Union[str, None], optional): Name of notebook to
                    create kernel pool.
                    Defaults to None.
                """
                delay = delay if delay is not None else self.fill_delay
                try:
                    loop = asyncio.get_event_loop()
                except RuntimeError:
                    loop = asyncio.new_event_loop()
                    asyncio.set_event_loop(loop)
                default_config: dict = self.kernel_pools_config.get("default", {})
                notebook_config: dict = self.kernel_pools_config.get(
                    notebook_name, default_config
                )
                kernel_env_variables: dict = notebook_config.get(
                    "kernel_env_variables",
                    default_config.get("kernel_env_variables", {}),
                )
                kernel_size: int = notebook_config.get(
                    "pool_size", default_config.get("pool_size", 1)
                )
                pool = self._pools.get(notebook_name, [])
                self._pools[notebook_name] = pool
                if "path" not in kwargs:
                    kwargs["path"] = (
                        os.path.dirname(notebook_name)
                        if notebook_name is not None
                        else self.root_dir
                    )
                kernel_env = os.environ.copy()
                kernel_env_arg = kwargs.get("env", {})
                kernel_env.update(kernel_env_arg)

                for key in kernel_env_variables:
                    if key not in kernel_env:
                        kernel_env[key] = kernel_env_variables[key]

                kernel_env.update(extra_kernel_env_variables)
                kernel_env[ENV_VARIABLE.VOILA_BASE_URL] = self.parent.base_url
                kernel_env[ENV_VARIABLE.VOILA_SERVER_URL] = self.parent.server_url
                kernel_env[ENV_VARIABLE.VOILA_APP_PORT] = str(self.parent.port)
                kernel_env[ENV_VARIABLE.VOILA_PREHEAT] = "True"
                kwargs["env"] = kernel_env

                heated = len(pool)

                def task_counter(tk):
                    nonlocal heated
                    heated += 1
                    if heated == kernel_size:
                        self.log.info(
                            "Kernel pool of %s is filled with %s kernel(s)",
                            notebook_name,
                            kernel_size,
                        )

                for _ in range(kernel_size - len(pool)):
                    # Start the work on the loop immediately, so it is ready when needed:
                    task = loop.create_task(
                        wait_before(
                            delay,
                            self._initialize(notebook_name, None, **kwargs),
                        )
                    )
                    pool.append(task)
                    task.add_done_callback(task_counter)

            async def restart_kernel(self, kernel_id: str, **kwargs) -> None:
                await ensure_async(super().restart_kernel(kernel_id, **kwargs))
                notebook_name = self._get_notebook_from_kernel(kernel_id)
                if notebook_name is not None:
                    await self._initialize(notebook_name, kernel_id, **kwargs)

            async def shutdown_kernel(self, kernel_id: str, *args, **kwargs):
                for pool in self._pools.values():
                    for i, f in enumerate(pool):
                        if f.done() and f.result().get("kernel_id") == kernel_id:
                            pool.pop(i)
                            break
                    else:
                        continue
                    break
                for value in self.notebook_data.values():
                    if kernel_id in value["kernel_ids"]:
                        value["kernel_ids"].remove(kernel_id)

                return await ensure_async(
                    super().shutdown_kernel(kernel_id, *args, **kwargs)
                )

            async def shutdown_all(self, *args, **kwargs):
                await ensure_async(super().shutdown_all(*args, **kwargs))
                # Parent doesn't correctly add all created kernels until they have completed startup:
                pools = self._pools
                self._pools = {}

                for value in self.notebook_data.values():
                    value["kernel_ids"] = set()

                for pool in pools.values():
                    # The iteration gets confused if we don't copy pool
                    for task in tuple(pool):
                        try:
                            result = await task
                        except Exception:
                            pass
                        else:
                            kid = result["kernel_id"]
                            if kid in self:
                                try:
                                    await ensure_async(
                                        self.shutdown_kernel(kid, *args, **kwargs)
                                    )
                                except RuntimeError:
                                    pass  # Kernel is not running

            async def _initialize(
                self, notebook_path: str, kernel_id: str = None, **kwargs
            ) -> str:
                """Run any configured initialization code in the kernel"""

                renderer = self._notebook_renderer_factory(notebook_path)
                await renderer.initialize()
                kernel_name = renderer.notebook.metadata.kernelspec.name

                if kernel_id is None:
                    kernel_id = await super().start_kernel(
                        kernel_name=kernel_name, **kwargs
                    )

                if renderer.notebook_path not in self.notebook_data:
                    self.notebook_data[renderer.notebook_path] = {
                        "notebook": renderer.notebook,
                        "template": renderer.template_name,
                        "theme": renderer.theme,
                        "kernel_name": kernel_name,
                        "kernel_ids": {kernel_id},
                    }
                else:
                    self.notebook_data[renderer.notebook_path]["kernel_ids"].add(
                        kernel_id
                    )

                kernel_future = self.get_kernel(kernel_id)
                task = asyncio.get_event_loop().create_task(
                    renderer.generate_content_hybrid(kernel_id, kernel_future)
                )
                return {"task": task, "renderer": renderer, "kernel_id": kernel_id}

            async def cull_kernel_if_idle(self, kernel_id: str):
                """Ensure we don't cull pooled kernels:
                (this logic assumes the init time is shorter than the cull time)
                """

                for pool in self._pools.values():
                    for i, f in enumerate(pool):
                        try:
                            if f.done() and f.result().get("kernel_id") == kernel_id:
                                return
                        except Exception:
                            pool.pop(i)
                return await ensure_async(super().cull_kernel_if_idle(kernel_id))

            def _notebook_renderer_factory(
                self, notebook_path: Union[str, None] = None
            ) -> NotebookRenderer:
                """Helper function to create `NotebookRenderer` instance.

                Args:
                    - notebook_path (Union[str, None], optional): Path to the
                    notebook. Defaults to None.
                """
                voila_configuration = self.parent.voila_configuration
                settings = self.parent.app.settings
                mathjax_config = settings.get("mathjax_config")
                mathjax_url = settings.get("mathjax_url")
                return NotebookRenderer(
                    voila_configuration=voila_configuration,
                    traitlet_config=self.parent.config,
                    notebook_path=notebook_path,
                    template_paths=self.parent.template_paths,
                    config_manager=self.parent.config_manager,
                    contents_manager=self.parent.contents_manager,
                    base_url=self.parent.base_url,
                    kernel_spec_manager=self.parent.kernel_spec_manager,
                    page_config=get_page_config(
                        base_url=self.parent.base_url,
                        settings=self.parent.app.settings,
                        log=self.parent.log,
                        voila_configuration=voila_configuration,
                    ),
                    mathjax_config=mathjax_config,
                    mathjax_url=mathjax_url,
                )

            def _notebook_filter(self, nb_path: Path) -> bool:
                """Helper to filter denylisted notebooks.

                Args:
                    nb_path (Path): Path to notebook

                Returns:
                    bool: return `False` if notebook is in `ipynb_checkpoints` folder or
                    is denylisted, `True` otherwise.
                """
                nb_name = str(nb_path)
                if ".ipynb_checkpoints" in nb_name:
                    return False
                for nb_pattern in self.preheat_denylist:
                    pattern = re.compile(nb_pattern)
                    if (nb_pattern in nb_name) or bool(pattern.match(nb_name)):
                        return False
                return True

            def _get_notebook_from_kernel(self, kernel_id: str) -> Union[None, str]:
                """Helper to get notebook name from heated kernel id.

                Args:
                    kernel_id (str): Kernel id

                Returns:
                    Union[None, str]: return associated notebook with kernel id.

                """
                for nb_name, data in self.notebook_data.items():
                    if kernel_id in data["kernel_ids"]:
                        return nb_name
                return None

    return VoilaKernelManager



import os

import pytest

BASE_DIR = os.path.dirname(__file__)


@pytest.fixture
def jupyter_server_args_extra():
    return ["--VoilaConfiguration.strip_sources=False"]


async def test_hello_world(http_server_client, print_notebook_url):
    response = await http_server_client.fetch(print_notebook_url)
    assert response.code == 200
    html_text = response.body.decode("utf-8")
    assert "Hi Voilà" in html_text
    assert "print" in html_text, "the source code should *NOT* be stripped"



#############################################################################
# Copyright (c) 2018, Voilà Contributors                                    #
# Copyright (c) 2018, QuantStack                                            #
#                                                                           #
# Distributed under the terms of the BSD 3-Clause License.                  #
#                                                                           #
# The full license is in the file LICENSE, distributed with this software.  #
#############################################################################

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import re
from collections import namedtuple

# Use "hatch version xx.yy.zz" to handle version changes
__version__ = "0.5.7"

# PEP440 version parser
_version_regex = re.compile(
    r"""
  (?P<major>\d+)
  \.
  (?P<minor>\d+)
  \.
  (?P<micro>\d+)
  (?P<releaselevel>((a|b|rc|\.dev)))?
  (?P<serial>\d+)?
  """,
    re.VERBOSE,
)

_version_fields = _version_regex.match(__version__).groupdict()  # type:ignore

VersionInfo = namedtuple(
    "VersionInfo", ["major", "minor", "micro", "releaselevel", "serial"]
)

version_info = VersionInfo(
    *[
        int(_version_fields["major"]),
        int(_version_fields["minor"]),
        int(_version_fields["micro"]),
        _version_fields["releaselevel"] or "",
        _version_fields["serial"] or "",
    ]
)



import os

import pytest

import voila.app

BASE_DIR = os.path.dirname(__file__)


class VoilaTest(voila.app.Voila):
    def listen(self):
        pass  # the ioloop is taken care of by the pytest-tornado framework


@pytest.fixture
def voila_config():
    return lambda app: None


@pytest.fixture
def voila_args_extra():
    return ["--VoilaExecutor.timeout=240"]


@pytest.fixture
def voila_config_file_paths_arg():
    # we don't want the tests to use any configuration on the system
    return "--VoilaTest.config_file_paths=[]"


@pytest.fixture
def voila_args(voila_notebook, voila_args_extra, voila_config_file_paths_arg):
    debug_args = (
        ["--VoilaTest.log_level=DEBUG"]
        if os.environ.get("VOILA_TEST_DEBUG", False)
        else []
    )
    return [voila_notebook, voila_config_file_paths_arg, *voila_args_extra, *debug_args]


@pytest.fixture
def preheat_mode():
    """Fixture used to activate/deactivate pre-heat kernel mode.
    Override this fixture in test file if you want to activate
    preheat mode.
    """
    return False


@pytest.fixture
def preheat_config(preheat_mode):
    return f"--preheat_kernel={preheat_mode}"


@pytest.fixture
def voila_app(voila_args, voila_config, preheat_config):
    voila_app = VoilaTest.instance()
    voila_app.initialize([*voila_args, "--no-browser", preheat_config])
    voila_config(voila_app)
    voila_app.start()
    yield voila_app
    voila_app.stop()
    voila_app.clear_instance()


@pytest.fixture
def app(voila_app):
    return voila_app.app



# tests programmatic config of template system
import os

import pytest

BASE_DIR = os.path.dirname(__file__)


@pytest.fixture
def voila_config():
    def config(app):
        pass

    os.environ["JUPYTER_CONFIG_DIR"] = os.path.join(BASE_DIR, "../configs/general")
    yield config
    del os.environ["JUPYTER_CONFIG_DIR"]


@pytest.fixture
def voila_config_file_paths_arg():
    # we don't want the tests to use any configuration on the system
    path = os.path.abspath(os.path.join(BASE_DIR, "../configs/general"))
    return "--VoilaTest.config_file_paths=[%r]" % path


@pytest.mark.skip(reason="TODO: update for JupyterLab extensions")
async def test_lists_extension(http_server_client, base_url):
    response = await http_server_client.fetch(base_url)
    assert response.code == 200
    html_text = response.body.decode("utf-8")
    assert "Hi Voilà" in html_text
    assert "ipytest/extension.js" in html_text



import asyncio
import os
import time

import pytest

BASE_DIR = os.path.dirname(__file__)
NOTEBOOK_EXECUTION_TIME = 3
NUMBER_PREHEATED_KERNEL = 2
TIME_THRESHOLD = 1


@pytest.fixture
def voila_config_file_paths_arg():
    path = os.path.join(BASE_DIR, "..", "configs", "preheat")
    return "--VoilaTest.config_file_paths=[%r]" % path


@pytest.fixture
def preheat_mode():
    return True


@pytest.fixture
def voila_notebook(notebook_directory):
    return os.path.join(notebook_directory, "preheat")


async def send_request(sc, url, wait=0):
    await asyncio.sleep(wait)
    real_time = time.time()
    response = await sc.fetch(url)
    real_time = time.time() - real_time
    html_text = response.body.decode("utf-8")
    return real_time, html_text


async def test_render_notebook_with_heated_kernel(http_server_client, base_url):
    await asyncio.sleep(NUMBER_PREHEATED_KERNEL * NOTEBOOK_EXECUTION_TIME + 1)
    time, text = await send_request(
        sc=http_server_client, url=f"{base_url}voila/render/pre_heat.ipynb"
    )

    assert "hello world" in text
    assert time < TIME_THRESHOLD
    await asyncio.sleep(NOTEBOOK_EXECUTION_TIME + 1)


async def test_render_denylisted_notebook_with_nornal_kernel(
    http_server_client, base_url
):
    await asyncio.sleep(NUMBER_PREHEATED_KERNEL * NOTEBOOK_EXECUTION_TIME + 1)
    time, text = await send_request(
        sc=http_server_client, url=f"{base_url}voila/render/denylisted.ipynb"
    )

    assert "hello world" in text
    assert time > TIME_THRESHOLD
    await asyncio.sleep(NOTEBOOK_EXECUTION_TIME + 1)



import asyncio
import os

import pytest


@pytest.fixture()
def voila_args_extra():
    return ['--VoilaKernelManager.default_env_variables={"FOO": "BAR"}']


@pytest.fixture
def preheat_mode():
    return True


@pytest.fixture
def voila_notebook(notebook_directory):
    return os.path.join(notebook_directory, "preheat", "default_env_variables.ipynb")


NOTEBOOK_EXECUTION_TIME = 2


async def send_request(sc, url, wait=0):
    await asyncio.sleep(wait)
    response = await sc.fetch(url)
    return response.body.decode("utf-8")


async def test_default_kernel_env_variable(http_server_client, base_url):
    html_text = await send_request(
        sc=http_server_client, url=base_url, wait=NOTEBOOK_EXECUTION_TIME + 1
    )
    assert "BAR" in html_text



from typing import Dict, List, Set
from voila.utils import filter_extension


federated_extensions = [{"name": "foo"}, {"name": "bar"}, {"name": "foo-bar"}]


def list_to_set(inp: List[Dict]) -> Set:
    return set([x["name"] for x in inp])


def test_filter_extension():
    extensions = filter_extension(federated_extensions=federated_extensions)
    assert list_to_set(extensions) == list_to_set(federated_extensions)


def test_filter_extension_with_disabled():
    disabled_extensions = ["bar"]
    extensions = filter_extension(
        federated_extensions=federated_extensions,
        disabled_extensions=disabled_extensions,
    )
    assert list_to_set(extensions) == set(["foo", "foo-bar"])



import os

import pytest
from jupyter_server.serverapp import ServerApp
from tornado import httpserver

BASE_DIR = os.path.dirname(__file__)


@pytest.fixture
def jupyter_server_config():
    return lambda app: None


@pytest.fixture
def jupyter_server_args_extra():
    return []


@pytest.fixture
def jupyter_server_args(notebook_directory, jupyter_server_args_extra):
    debug_args = (
        ["--ServerApp.log_level=DEBUG"]
        if os.environ.get("VOILA_TEST_DEBUG", False)
        else []
    )
    default_args = ["--ServerApp.token="]
    return [notebook_directory, *jupyter_server_args_extra, *debug_args, *default_args]


@pytest.fixture
def jupyter_server_app(jupyter_server_args, jupyter_server_config):
    jupyter_server_app = ServerApp.instance()
    # we monkey patch
    old_listen = httpserver.HTTPServer.listen
    httpserver.HTTPServer.listen = lambda *x, **y: None
    # NOTE: in voila's conftest.py we call config after initialize
    jupyter_server_config(jupyter_server_app)
    jupyter_server_app.initialize(jupyter_server_args)
    yield jupyter_server_app
    httpserver.HTTPServer.listen = old_listen
    ServerApp.clear_instance()


@pytest.fixture
def app(jupyter_server_app):
    return jupyter_server_app.web_app



# test all objects that should be configurable
import os

import pytest

BASE_DIR = os.path.dirname(__file__)


@pytest.fixture
def voila_config_file_paths_arg():
    path = os.path.join(BASE_DIR, "..", "configs", "general")
    return "--VoilaTest.config_file_paths=[%r]" % path


def test_config_app(voila_app):
    assert voila_app.voila_configuration.template == "test_template"


def test_config_kernel_manager(voila_app):
    assert voila_app.kernel_manager.cull_interval == 10


def test_config_contents_manager(voila_app):
    assert voila_app.contents_manager.use_atomic_writing is False


async def test_template(http_server_client, base_url):
    response = await http_server_client.fetch(base_url)
    assert response.code == 200
    assert "test_template.css" in response.body.decode("utf-8")
    assert "Hi Voilà" in response.body.decode("utf-8")



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.import click

# Heavily inspired by:
# - https://github.com/jupyterlab/jupyterlab/blob/master/buildutils/src/bumpversion.ts
# - https://github.com/jupyterlab/retrolab/blob/main/buildutils/src/release-bump.ts

import click
from jupyter_releaser.util import get_version, is_prerelease, run

OPTIONS = ["major", "minor", "release", "build", "patch", "next"]


def patch(force=False):
    version = get_version()
    if is_prerelease(version):
        raise Exception("Can only make a patch release from a final version")

    run("hatch version patch", quiet=True)

    # Version the changed
    cmd = "jlpm && jlpm run lerna version patch --no-push --force-publish --no-git-tag-version"
    if force:
        cmd += " --yes"
    run(cmd)


def update(spec, force=False):
    prev = get_version()

    is_final = not is_prerelease(prev)

    if is_final and spec == "build":
        raise Exception("Cannot increment a build on a final release")

    # If this is a major release during the alpha cycle, bump
    # just the Python version.
    if "a" in prev and spec == "major":
        run(f"hatch version {spec}")
        return

    # Determine the version spec to use for lerna and hatch
    py_spec = spec
    lerna_version = "preminor"
    if spec == "build":
        lerna_version = "prerelease"
        if "a" in prev:
            py_spec = "a"
        elif "b" in prev:
            py_spec = "b"
        elif "rc" in prev:
            py_spec = "rc"
    # a -> b
    elif spec == "release" and "a" in prev:
        lerna_version = "prerelease --preid=beta"
        py_spec = "beta"
    # b -> rc
    elif spec == "release" and "b" in prev:
        lerna_version = "prerelease --preid=rc"
        py_spec = "rc"
    # rc -> final
    elif spec == "release" and "c" in prev:
        lerna_version = "patch"
        py_spec = "release"
    elif spec == "release":
        py_spec = "minor,alpha"

    if lerna_version == "preminor":
        lerna_version += " --preid=alpha"

    cmd = f"jlpm && jlpm run lerna version --force-publish --no-push --no-git-tag-version {lerna_version}"
    if force:
        cmd += " --yes"

    # For a preminor release, we bump 10 minor versions so that we do
    # not conflict with versions during minor releases of the top level package.
    if lerna_version == "preminor":
        for _i in range(10):
            run(cmd)
    else:
        run(cmd)

    # Bump the Python version
    run(f"hatch version {py_spec}")


@click.command()
@click.option("--force", default=False, is_flag=True)
@click.argument("spec", nargs=1)
def bump(force, spec):
    status = run("git status --porcelain").strip()
    if len(status) > 0:
        raise Exception("Must be in a clean git state with no untracked files")

    # Make sure we have a valid version spec.
    if spec not in OPTIONS:
        raise ValueError(f"Version spec must be one of: {OPTIONS}")

    prev = get_version()
    is_final = not is_prerelease(prev)
    if spec == "next":
        spec = "patch" if is_final else "build"

    if spec == "patch":
        patch(force)
        return

    update(spec, force)


if __name__ == "__main__":
    bump()
